/*  A quick hack version of the scoreboard  */
#include <osfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <math.h>

#include "packet.h"
#include "scoreboard.h"

#define ASSERT(x) if (!(x)) {printf ("Assert SB failed\n"); exit(1);}

#define SBNI SBN[i%SBSIZE]

// last_ack = TCP last ack
void ScoreBoard::UpdateScoreBoard (int last_ack, Packet *pkt, int &retran_data)
{
	int i, sack_index, sack_left, sack_right;

	//  If there is no scoreboard, create one.
	if (length_ == 0) {
		i = last_ack+1;
		SBNI.seq_no_ = i;
		SBNI.ack_flag_ = 0;
		SBNI.sack_flag_ = 0;
		SBNI.retran_ = 0;
		SBNI.snd_nxt_ = 0;
		first_ = i%SBSIZE;
		length_++;
		if (length_ >= SBSIZE) {
			printf ("Error, scoreboard too large\n");
			exit(1);
		}
	}	

	for (sack_index=0; sack_index < pkt->sack_length_; sack_index++) {
		sack_left = pkt->sa_left_[sack_index];
		sack_right = pkt->sa_right_[sack_index];

		//  Create new entries off the right side.
		if (sack_right > SBN[(first_+length_)%SBSIZE].seq_no_) {
			//  Create new entries
			for (i = SBN[(first_+length_+SBSIZE-1)%SBSIZE].seq_no_+1; i<sack_right; i++) {
				SBNI.seq_no_ = i;
				SBNI.ack_flag_ = 0;
				SBNI.sack_flag_ = 0;
				SBNI.retran_ = 0;
				SBNI.snd_nxt_ = 0;
				length_++;
				if (length_ >= SBSIZE) {
					printf ("Error, scoreboard too large\n");
					exit(1);
				}
			}
		}

		//  Advance the left edge of the block.
		if (SBN[first_].seq_no_ <= last_ack) {
			for (i=SBN[(first_)%SBSIZE].seq_no_; i<=last_ack; i++) {
				//  Advance the ACK
				if (SBNI.seq_no_ <= last_ack) {
					ASSERT(first_ == i%SBSIZE);
					first_ = (first_+1)%SBSIZE; 
					length_--;
					ASSERT(length_ >= 0);

					SBNI.ack_flag_ = 1;
					SBNI.sack_flag_ = 1;
					if (SBNI.retran_) {
						retran_data--;
						SBNI.retran_ = 0;
						ASSERT(retran_data >= 0);
						SBNI.snd_nxt_ = 0;
					}
					if (length_==0) 
					  break;
				}
			}
		}
		
		for (i=SBN[(first_)%SBSIZE].seq_no_; i<sack_right; i++) {
			//  If the segment was retransmitted, check to see if it has 
			//  been dropped.
			if (SBNI.retran_) {
				if (SBNI.snd_nxt_ < sack_right) {
					SBNI.retran_ = 0;
					SBNI.snd_nxt_ = 0;
					retran_data--;
					ASSERT(retran_data >= 0);
				}
			}

			//  Check to see if this segment is now covered by the sack block
			if (SBNI.seq_no_ >= sack_left && SBNI.seq_no_ < sack_right) {
				if (! SBNI.sack_flag_) {
					//  If it was retransmitted, adjust retran_data:
					if (SBNI.retran_) {
						SBNI.retran_ = 0;
						SBNI.snd_nxt_ = 0;
						retran_data--;
						ASSERT(retran_data >= 0);
					}
					SBNI.sack_flag_ = 1;
				}
			}
		}
	}
}

int ScoreBoard::GetNextRetran()	    // Returns sequence number of next pkt...
{
	int i;

	for (i=SBN[(first_)%SBSIZE].seq_no_; i<SBN[(first_)%SBSIZE].seq_no_+length_; i++) {
		if (!SBNI.ack_flag_ && !SBNI.sack_flag_ && !SBNI.retran_) {
			return (i);
		}
	}
	return (-1);
}


void ScoreBoard::MarkRetran (int retran_seqno, int snd_nxt)
{
	SBN[retran_seqno%SBSIZE].retran_ = 1;
	SBN[retran_seqno%SBSIZE].snd_nxt_ = snd_nxt;
}

void ScoreBoard::ClearRetran (int &retran_data)
{
	int i;

	for (i=SBN[(first_)%SBSIZE].seq_no_; i<SBN[(first_)%SBSIZE].seq_no_+length_; i++) {
		if (SBNI.retran_) {
			SBNI.retran_ = 0;
			SBNI.snd_nxt_ = 0;
			retran_data--;
			ASSERT(retran_data >= 0);
		}
	}
}

int ScoreBoard::LastElementOrLastAck(int last_ack)
{
        if (length_==0) {
		return (last_ack);
	} 
	else {
		return (SBN[(first_+length_-1)%SBSIZE].seq_no_);
	}
}


