/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: tcp-sack1.cc,v 1.3 95/12/11 18:41:59 kfall Exp $ (LBL)";
#endif

#include <stdio.h>
#include <sys/types.h>

#include "tcp.h"
#include "scoreboard.h"
#include "random.h"
#include "trace.h"

#define INFINITY 999999

class Sack1TcpAgent : public TcpAgent {
 public:
	Sack1TcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	void plot();
	void send(int force, int reason, int maxburst);
 protected:
	u_int dupwnd_;
	int retran_data_;  /*  For future use  */
	ScoreBoard scb_;
};

static class Sack1TcpMatcher : public Matcher {
public:
	Sack1TcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-sack1") == 0)
			return (new Sack1TcpAgent());
		return (0);
	}
} matcher_sack;

int Sack1TcpAgent::window()
{
	int win = int(cwnd_) + dupwnd_;
	if (win > int(2 * wnd_))
		win = int(2 * wnd_);
	return (win);
}

Sack1TcpAgent::Sack1TcpAgent() : dupwnd_(0)
{
}

void Sack1TcpAgent::recv(Packet *pkt)
{
	int xmit_seqno;

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
//	printf ("last_ack = %d, recover = %d, ack = %d", last_ack_, recover_, pkt->seqno_);
	if (pkt->seqno_ > last_ack_ && 
	    ((recover_ == 0) || (pkt->seqno_ >= recover_))) {
		dupwnd_ = 0;
		newack(pkt);
                opencwnd();
		scb_.ClearScoreBoard();
		recover_ = 0;
   	} 
	else if (pkt->seqno_ == last_ack_ ||                      /* dupack  */
		 (dupacks_ > 0) && (pkt->seqno_ < recover_))  {   /* or cong */
		scb_.UpdateScoreBoard (last_ack_, pkt, retran_data_);
		newtimer(pkt);

		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
			dupwnd_ = NUMDUPACKS;
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			closecwnd(1);
			recover_ = maxseq_;
			recover_cause_ = 1;
			reset_rtx_timer(1);
			
			xmit_seqno = scb_.GetNextRetran ();
			if (xmit_seqno == -1) {
				printf ("Sack rxmit: A case was reached which should never happen\n");
				xmit_seqno = seqno_++;
			} else {
			        /*  Should do this?, but reno doesn't:
				    dupwnd_--;  */
			        retran_data_++;
				scb_.MarkRetran (xmit_seqno, INFINITY);
			}
                        output(xmit_seqno, REASON_DUPACK);
		} else if (dupacks_ > NUMDUPACKS) {
			if (last_ack_ == pkt->seqno_) {
				++dupwnd_;
			}
		        else {
				last_ack_ = pkt->seqno_;
				highest_ack_ = last_ack_;
			}
		}
	} else {
		printf ("Sack agent: A case was reached which should never happen\n");
	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0 || dupacks_ > NUMDUPACKS - 1)
		send(0, REASON_NORMAL, maxburst_);
}

void Sack1TcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
		TcpAgent::timeout(tno);
	}
}

void Sack1TcpAgent::send(int force, int reason, int maxburst)
{
        int win = window();
        int npackets = 0;
	int xmit_seqno;

//	printf ("seqno = %d    win = %d   xmits: ", seqno_, win);
        while (seqno_ <= highest_ack_ + win && seqno_ < curseq_) {
                if (overhead_ == 0 || force) {
			xmit_seqno = scb_.GetNextRetran ();
			if (xmit_seqno == -1) {
				xmit_seqno = seqno_++;
			}
			else {
			        retran_data_++;
				dupwnd_--;
				scb_.MarkRetran (xmit_seqno, INFINITY);
				win = window();
			}
//			printf ("%d ", xmit_seqno);
                        output(xmit_seqno, reason);
                        npackets++;
                } else if (!pending_[TCP_TIMER_DELSND]) {
                        /*
                         * Set a delayed send timeout.
                         */
                        sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
                        return;
                }
                if (maxburst && npackets == maxburst)
                        break;
        }
//	printf ("\n");
}

void Sack1TcpAgent::plot()
{
	double t = Scheduler::instance().clock();
	sprintf(trace_->buffer(), "t %g %d rtt %g\n", t, class_, t_rtt_ * TCP_TICK);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d dev %g\n", t, class_, t_rttvar_ * TCP_TICK);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d win %f\n", t, class_, cwnd_+dupwnd_);
	trace_->dump();
	sprintf(trace_->buffer(), "t %g %d bck %d\n", t, class_, t_backoff_);
	trace_->dump();
}
